/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/* Native JNI implementation of anaglyph creation for speed.  */

#include "eu_domob_anacam_Anaglyph.h"
#include "eu_domob_anacam_MyBitmap.h"
#include "eu_domob_anacam_YuvDecoder.h"

#include <cassert>
#include <cstddef>
#include <cstdint>
#include <cstdio>
#include <cstring>

//#undef NDEBUG

#undef assert
#ifdef NDEBUG
  #define assert(cond)
#else /* NDEBUG.  */
  #define assert(cond) \
    if (!(cond)) \
      { \
        FILE* f = fopen ("/sdcard/anacam.log", "a"); \
        fprintf (f, "Assert failed: %s\n", #cond); \
        fclose (f); \
      }
#endif /* NDEBUG.  */

/* ************************************************************************** */
/* Anaglyph.  */

/**
 * Helper routine that does multiply-blending.
 * @param a First pixel value.
 * @param b Second pixel value.
 * @return Value of both multiplied together.
 */
static jint
multiply (jint a, jint b)
{
  jint res = 0xFF000000;
  for (unsigned channel = 0; channel < 3; ++channel)
    {
      const unsigned shift = 8 * channel;
      const jint mask = (0xFF << shift);
      const jint compA = ((a & mask) >> shift);
      const jint compB = ((b & mask) >> shift);
      assert (compA >= 0 && compA <= 255);
      assert (compB >= 0 && compB <= 255);
      const jint compRes = compA * compB / 0xFF;
      assert ((compRes & ~0xFF) == 0);
      res |= (compRes << shift);
    }

  return res;
}

/* Implement the native anaglyph creation method.  */
JNIEXPORT void JNICALL
Java_eu_domob_anacam_Anaglyph_doCombine (JNIEnv* env, jclass clazz,
                                         jint lcol, jint rcol,
                                         jlong limg, jlong rimg,
                                         jintArray dest)
{
  const jsize len = env->GetArrayLength (dest);

  jint* lpx = reinterpret_cast<jint*> (limg);
  jint* rpx = reinterpret_cast<jint*> (rimg);

  jint* dpx = env->GetIntArrayElements (dest, NULL);

  /* Check the assumption made below that the screening colours
     only have 0 and 255 as intensities.  */
#ifndef NDEBUG
  for (unsigned channel = 0; channel < 4; ++channel)
    {
      const unsigned shift = 8 * channel;
      const uint32_t mask = (0xFFu << shift);
      const uint32_t l = ((lcol & mask) >> shift);
      const uint32_t r = ((rcol & mask) >> shift);
      assert (l == 0xFFu || l == 0x00u);
      assert (r == 0xFFu || r == 0x00u);
    }
#endif /* NDEBUG.  */

  /* We assume that the screening colours have only 0 and 255 intensities.
     This is checked above.  Then the screening can be expressed without
     multiplication, simply as bitwise OR.

     If the colours are *exactly* complementary (like red-cyan), then
     we can further simplify the calculation:  Then for each channel
     always one of the screened pixels stays untouched (where screening
     colour is 0), and the other gets set to 255 (where
     screening colour is 255).  Then their product is simply the value
     of the untouched pixel.  */
  const bool complementary = ((rcol & ~0xFF000000) == ~lcol);

  if (complementary)
    for (jsize i = 0; i < len; ++i)
      dpx[i] = ((lpx[i] & ~lcol) | (rpx[i] & ~rcol) | 0xFF000000);
  else
    for (jsize i = 0; i < len; ++i)
      {
        const jint lscr = (lpx[i] | lcol);
        const jint rscr = (rpx[i] | rcol);
        dpx[i] = multiply (lscr, rscr);
      }

  env->ReleaseIntArrayElements (dest, dpx, 0);
}

/* ************************************************************************** */
/* MyBitmap.  */

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_freeNative (JNIEnv* env, jclass clazz, jlong ptr)
{
  jint* cPtr = reinterpret_cast<jint*> (ptr);
  assert (cPtr);
  delete[] cPtr;
}

JNIEXPORT jlong JNICALL
Java_eu_domob_anacam_MyBitmap_copyToNative (JNIEnv* env, jclass clazz,
                                            jintArray data)
{
  const jsize len = env->GetArrayLength (data);
  jint* jPtr = env->GetIntArrayElements (data, NULL);

  jint* cPtr = new jint[len];
  std::memcpy (cPtr, jPtr, sizeof (jint) * len);

  env->ReleaseIntArrayElements (data, jPtr, 0);

  return reinterpret_cast<jlong> (cPtr);
}

JNIEXPORT void JNICALL
Java_eu_domob_anacam_MyBitmap_doGrayscale (JNIEnv* env, jclass clazz,
                                           jintArray data)
{
  const jsize len = env->GetArrayLength (data);
  jint* ptr = env->GetIntArrayElements (data, NULL);

  /* Use relative luminances of the colour channels from:
     http://stackoverflow.com/questions/8025621/how-to-convert-rgb565-to-yuv420sp-faster-on-android.  */
  static const jint Y[] = {19595, 38470, 7471};
  static const unsigned SHIFT = 16;

  for (jsize i = 0; i < len; ++i)
    {
      const jint cur = ptr[i];
      jint res = 0;
      for (unsigned channel = 0; channel < 3; ++channel)
        {
          const unsigned shift = 8 * channel;
          const jint mask = (0xFF << shift);
          const jint comp = ((cur & mask) >> shift);
          assert (comp >= 0 && comp <= 255);
          res += comp * Y[channel];
        }
      res >>= SHIFT;
      assert (res >= 0 && res <= 255);

      jint combined = 0xFF000000;
      for (unsigned channel = 0; channel < 3; ++channel)
        {
          const unsigned shift = 8 * channel;
          combined |= (res << shift);
        }

      ptr[i] = combined;
    }

  env->ReleaseIntArrayElements (data, ptr, 0);
}

/* ************************************************************************** */
/* YuvDecoder.  */

/* Implement native YUV decoding.  */
JNIEXPORT void JNICALL
Java_eu_domob_anacam_YuvDecoder_decodeYUV420SP
  (JNIEnv* env, jclass clazz, jintArray javaRGB, jbyteArray javaYUV,
   jint width, jint height, jboolean gray)
{
  /* This code is from:
     https://code.google.com/p/android/issues/detail?id=22966.  */

  jint* rgb = env->GetIntArrayElements (javaRGB, NULL);
  jbyte* yuv = env->GetByteArrayElements (javaYUV, NULL);

  const jint frameSize = width * height;
  for (jint j = 0, yp = 0; j < height; j++)
    {
      jint uvp = frameSize + (j >> 1) * width;
      jint u = 0, v = 0;
      for (jint i = 0; i < width; i++, yp++)
        {
          jint y = (0xFF & yuv[yp]) - 16;
          if (y < 0)
            y = 0;
          assert (y >= 0 && y <= 255);

          if (gray)
            {
              jint res = 0xFF000000;
              for (unsigned channel = 0; channel < 3; ++channel)
                {
                  const unsigned shift = 8 * channel;
                  res |= (y << shift);
                }
              rgb[yp] = res;
            }
          else
            {
              if ((i & 1) == 0)
                {
                  v = (0xFF & yuv[uvp++]) - 128;
                  u = (0xFF & yuv[uvp++]) - 128;
                }

              const jint y1192 = 1192 * y;
              jint r = (y1192 + 1634 * v);
              jint g = (y1192 - 833 * v - 400 * u);
              jint b = (y1192 + 2066 * u);
           
              if (r < 0)
                r = 0;
              else if (r > 262143)
                r = 262143;

              if (g < 0)
                g = 0;
              else if (g > 262143)
                g = 262143;

              if (b < 0)
                b = 0;
              else if (b > 262143)
                b = 262143;

              rgb[yp] = 0xFF000000 | ((r << 6) & 0xFF0000)
                                   | ((g >> 2) & 0xFF00) | ((b >> 10) & 0xFF);
            }
        } 
    }

  env->ReleaseIntArrayElements (javaRGB, rgb, 0);
  env->ReleaseByteArrayElements (javaYUV, yuv, 0);
}
