/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package eu.domob.anacam;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.app.ProgressDialog;

import android.content.Context;
import android.content.Intent;
import android.content.SharedPreferences;

import android.graphics.Bitmap;
import android.graphics.BitmapFactory;

import android.hardware.Camera;

import android.os.Bundle;
import android.os.Debug;
import android.os.Handler;
import android.os.Message;

import android.preference.PreferenceManager;

import android.text.method.LinkMovementMethod;

import android.util.Log;

import android.view.Menu;
import android.view.MenuInflater;
import android.view.MenuItem;
import android.view.View;
import android.view.Window;
import android.view.WindowManager;

import android.widget.FrameLayout;
import android.widget.TextView;
import android.widget.Toast;

import java.io.File;
import java.io.FileOutputStream;

/**
 * This is the main class for AnaCam.
 */
public class AnaCam extends Activity
  implements Camera.PictureCallback,
             View.OnClickListener, View.OnLongClickListener
{

  /** Logging tag.  */
  public static final String TAG = "AnaCam";

  /** No profiling.  */
  public static final int PROFILE_NONE = 0;
  /** Profile taking the first picture (corresponding previews).  */
  public static final int PROFILE_FIRST = 1;
  /** Profile taking the second picture (previews).  */
  public static final int PROFILE_SECOND = 2;
  /** Profile combining and saving final picture.  */
  public static final int PROFILE_SAVING = 3;
  /** Configure whether we want profiling or not.  */
  public static final int PROFILING = PROFILE_NONE;

  /** ID for help dialog.  */
  private static final int DIALOG_HELP = 0;
  /** ID for about dialog.  */
  private static final int DIALOG_ABOUT = 1;
  /** ID for progress dialog.  */
  private static final int DIALOG_PROGRESS = 2;

  /** Preview object.  */
  private Preview preview;

  /** SharedPreferences handler.  */
  private SharedPreferences pref;

  /** Remember here the jpeg data of a first picture we have.  */
  private byte[] firstJpeg;
  /** Bitmap of first picture.  */
  private MyBitmap firstBitmap;

  /**
   * Initialise members and the layout.
   * @param savedInstanceState Saved state.
   */
  @Override
  public void onCreate(Bundle savedInstanceState)
  {
    super.onCreate(savedInstanceState);
    pref = PreferenceManager.getDefaultSharedPreferences (this);

    requestWindowFeature (Window.FEATURE_NO_TITLE);
    getWindow ().setFlags (WindowManager.LayoutParams.FLAG_FULLSCREEN,
                           WindowManager.LayoutParams.FLAG_FULLSCREEN);
    setContentView(R.layout.main);

    preview = new Preview (this);
    FrameLayout container = (FrameLayout) findViewById (R.id.preview);
    container.addView (preview);

    firstBitmap = null;

    preview.setOnClickListener (this);
    preview.setOnLongClickListener (this);
  }

  /**
   * On click on the view, take picture.
   * @param v The view we clicked on.
   */
  public void onClick (View v)
  {
    assert (v == preview);
    preview.takePicture (this);
  }

  /**
   * On long click, cancel current anaglyph and start fresh.
   * @param v The view we clicked on.
   */
  public boolean onLongClick (View v)
  {
    assert (v == preview);
    if (firstBitmap == null)
      return false;

    firstBitmap = null;
    firstJpeg = null;
    preview.setFirstPicture (null);

    final Toast t = Toast.makeText (this, getString (R.string.cancel_msg),
                                    Toast.LENGTH_SHORT);
    t.show ();

    logMemory ("after cancelling current anaglyph");
    return true;
  }

  /**
   * Picture callback when a shot is taken.
   * @param data Picture data.
   * @param cam The camera used.
   */
  public void onPictureTaken (final byte[] data, Camera cam)
  {
    final Bitmap bmp = BitmapFactory.decodeByteArray (data, 0, data.length);

    /* If we don't yet have a first picture, simply store it and wait for
       the second one.  */
    if (firstBitmap == null)
      {
        preview.freeBuffers ();

        firstJpeg = data;
        logMemory ("before allocating firstBitmap");
        firstBitmap = new MyBitmap (bmp);

        preview.setFirstPicture (bmp);
        bmp.recycle ();

        logMemory ("before starting preview with first picture taken");
        preview.start ();
      }

    /* Otherwise, we have all now.  Generate the final anaglyph and
       write everything to disk.  */
    else
      {
        preview.setFirstPicture (null);
        preview.freeBuffers ();

        showDialog (DIALOG_PROGRESS);
        if (PROFILING == PROFILE_SAVING)
          Debug.startMethodTracing ();

        final int MSG_OK = 0;
        final int MSG_ERROR = 1;
        final int MSG_DISMISS = 2;

        final Context c = this;
        final Handler handler = new Handler ()
          {
            @Override
            public void handleMessage (Message msg)
            {
              switch (msg.what)
                {
                  case MSG_OK:
                    final Toast t = Toast.makeText (c, (String) msg.obj,
                                                    Toast.LENGTH_SHORT);
                    t.show ();
                    break;

                  case MSG_ERROR:
                    AlertDialog dlg = new AlertDialog.Builder (c).create ();
                    dlg.setTitle (getString (R.string.saveError_title));
                    dlg.setMessage ((String) msg.obj);
                    dlg.setIcon (android.R.drawable.ic_dialog_alert);
                    dlg.show ();
                    break;

                  case MSG_DISMISS:
                    if (PROFILING == PROFILE_SAVING)
                      Debug.stopMethodTracing ();

                    dismissDialog (DIALOG_PROGRESS);
                    assert (firstBitmap == null);
                    firstJpeg = null;
                    
                    logMemory ("before starting fresh preview");
                    preview.start ();
                    break;

                  default:
                    assert (false);
                    break;
                }
            }
          };

        Thread saver = new Thread ()
          {
            @Override
            public void run ()
            {
              final boolean gray = pref.getBoolean ("grayscale", false);

              Anaglyph anaBuilder = new Anaglyph (c);
              if (gray)
                firstBitmap.toGrayscale ();
              anaBuilder.setFirst (firstBitmap);
              firstBitmap = null;

              logMemory ("before allocating second bitmap");
              MyBitmap my = new MyBitmap (bmp);
              bmp.recycle ();
              if (gray)
                my.toGrayscale ();
              anaBuilder.setSecond (my);

              logMemory ("before combining anaglyph");
              my = anaBuilder.combine ();
              anaBuilder = null;

              logMemory ("before creating anaglyph bitmap");
              final Bitmap anaBitmap
                = Bitmap.createBitmap (my.getData (), my.width, my.height,
                                       Bitmap.Config.ARGB_8888);
              my = null;
              final int quality = preview.getJpegQuality ();

              String folder = getString (R.string.folder_default);
              folder = pref.getString ("folder", folder);
              String order = getString (R.string.order_default);
              order = pref.getString ("order", order);

              logMemory ("before saving");
              try
                {
                  String shortFile;
                  if (order.equals ("left"))
                    shortFile = saveFiles (firstJpeg, data, anaBitmap,
                                           folder, quality);
                  else
                    shortFile = saveFiles (data, firstJpeg, anaBitmap,
                                           folder, quality);
                  anaBitmap.recycle ();

                  String str = getString (R.string.saveSuccess);
                  str = String.format (str, shortFile);
                  handler.sendMessage (Message.obtain (handler, MSG_OK, str));
                }
              catch (Exception exc)
                {
                  final Message msg = Message.obtain (handler, MSG_ERROR,
                                                      exc.getMessage ());
                  handler.sendMessage (msg);
                }

              handler.sendEmptyMessage (MSG_DISMISS);
            }
          };
        saver.start ();
      }
  }

  /**
   * Create the menu.
   * @param menu The menu to create.
   * @return True for success.
   */
  @Override
  public boolean onCreateOptionsMenu (Menu menu)
  {
    MenuInflater inflater = getMenuInflater ();
    inflater.inflate (R.menu.main, menu);
    return true;
  }

  /**
   * Handle menu "click".
   * @param itm The menu item selected.
   * @return True if the event was handled.
   */
  @Override
  public boolean onOptionsItemSelected (MenuItem itm)
  {
    switch (itm.getItemId ())
      {
      case R.id.help:
        showDialog (DIALOG_HELP);
        return true;

      case R.id.about:
        showDialog (DIALOG_ABOUT);
        return true;

      case R.id.preferences:
        startActivity (new Intent (this, Preferences.class));
        return true;

      default:
        return super.onOptionsItemSelected (itm);
      }
  }

  /**
   * Create dialogs.
   * @param id Dialog to create.
   */
  @Override
  public Dialog onCreateDialog (int id)
  {
    if (id == DIALOG_PROGRESS)
      {
        ProgressDialog progress = new ProgressDialog (this);
        progress.setCancelable (false);
        progress.setProgressStyle (ProgressDialog.STYLE_SPINNER);
        progress.setMessage (getString (R.string.saving_msg));
        return progress;
      }

    Dialog dlg = new Dialog (this);

    switch (id)
      {
      case DIALOG_HELP:
        dlg.setContentView (R.layout.help);

        TextView tv = (TextView) dlg.findViewById (R.id.help_link);
        tv.setMovementMethod (LinkMovementMethod.getInstance ());

        dlg.setTitle (R.string.help_title);
        break;

      case DIALOG_ABOUT:
        dlg.setContentView (R.layout.about);

        tv = (TextView) dlg.findViewById (R.id.about_version);
        final String aboutVersion = getString (R.string.about_version);
        final String appName = getString (R.string.app_name);
        final String appVersion = getString (R.string.app_version);
        tv.setText (String.format (aboutVersion, appName, appVersion));

        tv = (TextView) dlg.findViewById (R.id.about_link1);
        tv.setMovementMethod (LinkMovementMethod.getInstance ());

        tv = (TextView) dlg.findViewById (R.id.about_link2);
        tv.setMovementMethod (LinkMovementMethod.getInstance ());

        dlg.setTitle (R.string.about_title);
        break;

      default:
        assert (false);
      }

    return dlg;
  }

  /**
   * Log memory usage.
   * @param note Note describing where we are in the process.
   */
  public static void logMemory (String note)
  {
    if (BuildConfig.DEBUG)
      {
        System.gc ();

        final Runtime rt = Runtime.getRuntime ();
        Log.d (TAG, String.format ("Memory usage %s:", note));
        Log.d (TAG, String.format ("  Max:    %d", rt.maxMemory ()));
        Log.d (TAG, String.format ("  Total:  %d", rt.totalMemory ()));
        Log.d (TAG, String.format ("  Free:   %d", rt.freeMemory ()));
        Log.d (TAG, String.format ("  Used:   %d",
                                   rt.totalMemory () - rt.freeMemory ()));
        Log.d (TAG, String.format ("  Native: %d",
                                   Debug.getNativeHeapAllocatedSize ()));
      }
  }

  /**
   * Handle all the file-system stuff of saving the data.
   * @param left JPEG byte data for left image.
   * @param right JPEG byte data for right image.
   * @param anaglyph Anaglyph combined into Bitmap object.
   * @param folder Folder where to save them to.
   * @param quality JPEG quality for saving.
   * @return Short name of saved file.
   * @throws Exception On error.
   */
  private String saveFiles (byte[] left, byte[] right, Bitmap anaglyph,
                            String folder, int quality)
    throws Exception
  {
    /* Ensure the directory is there and usable.  */
    final File dir = new File (folder);
    if (!dir.exists ())
      dir.mkdirs ();
    if (!dir.exists () || !dir.isDirectory ())
      {
        String msg = getString (R.string.error_noDir);
        msg = String.format (msg, folder);
        throw new RuntimeException (msg);
      }

    /* Look for the next still available number.  */
    int num = 0;
    File fLeft, fRight, fAna;
    while (true)
      {
        ++num;

        if (num >= 10000)
          throw new RuntimeException (getString (R.string.error_numsFull));

        String name;

        name = String.format (getString (R.string.file_left), num);
        fLeft = new File (dir, name);
        if (fLeft.exists ())
          continue;

        name = String.format (getString (R.string.file_right), num);
        fRight = new File (dir, name);
        if (fRight.exists ())
          continue;

        name = String.format (getString (R.string.file_anaglyph), num);
        fAna = new File (dir, name);
        if (fAna.exists ())
          continue;
        
        break;
      }

    /* Actually write.  */
    FileOutputStream out;
    out = new FileOutputStream (fLeft);
    out.write (left, 0, left.length);
    out.close ();
    out = new FileOutputStream (fRight);
    out.write (right, 0, right.length);
    out.close ();
    out = new FileOutputStream (fAna);
    anaglyph.compress (Bitmap.CompressFormat.JPEG, quality, out);
    out.close ();

    final String shrt = String.format (getString (R.string.file_short), num);
    Log.i (TAG, String.format ("Saved %s.", shrt));
    return shrt;
  }

}
