/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package eu.domob.anacam;

import android.content.Context;
import android.content.SharedPreferences;

import android.preference.PreferenceManager;

/**
 * This class does the calculation for combining images
 * into anaglyphs.  This class keeps the buffer for the output pixels
 * allocated to avoid allocation-overhead, which does not matter for the
 * previews anyway.
 */
public class Anaglyph
{

  /** Context used to query preferences and ressource strings.  */
  private Context context;
  /** Shared preferences object.  */
  private SharedPreferences pref;

  /** The first image.  */
  private MyBitmap first;
  /** The second image.  */
  private MyBitmap second;

  /** Buffer for combined pixels.  */
  private int[] res;

  /**
   * Construct it, with images left blank for now.
   * @param c Context to use.
   */
  public Anaglyph (Context c)
  {
    context = c;
    pref = PreferenceManager.getDefaultSharedPreferences (context);

    first = null;
    second = null;

    res = null;
  }

  /**
   * Set first image.
   * @param img New first image.
   */
  public void setFirst (MyBitmap img)
  {
    first = img;
    if (first != null)
      first.toNative ();
  }

  /**
   * Set second image.
   * @param img New second image.
   */
  public void setSecond (MyBitmap img)
  {
    second = img;
    if (second != null)
      second.toNative ();
  }

  /**
   * Free buffer to make memory available.
   */
  public void freeBuffers ()
  {
    res = null;
  }

  /**
   * Combine both images into an anaglyph.
   * @return The constructed anaglyph.
   */
  public MyBitmap combine ()
  {
    assert (first != null && second != null);
    final int w = first.width;
    final int h = first.height;
    if (second.width != w || second.height != h)
      throw new RuntimeException ("Image dimensions mismatch for anaglyph.");

    /* Use preferences to pick correct screening colours.  */
    int col1, col2;
    String type = context.getString (R.string.type_default);
    type = pref.getString ("type", type);
    String order = context.getString (R.string.order_default);
    order = pref.getString ("order", order);
    if (type.equals ("red-cyan"))
      {
        col1 = 0xFFFF0000;
        col2 = 0xFF00FFFF;
      }
    else if (type.equals ("red-green"))
      {
        col1 = 0xFFFF0000;
        col2 = 0xFF00FF00;
      }
    else if (type.equals ("red-blue"))
      {
        col1 = 0xFFFF0000;
        col2 = 0xFF0000FF;
      }
    else
      throw new RuntimeException ("Invalid type preference!");
    if (order.equals ("left"))
      {
        final int tmp = col1;
        col1 = col2;
        col2 = tmp;
      }
    else
      assert (order.equals ("right"));

    if (res == null || res.length != w * h)
      res = new int[w * h];
    doCombine (col1, col2, first.getNativeData (),
               second.getNativeData (), res);

    return new MyBitmap (res, w, h);
  }

  /**
   * Native method that does the number-crunching and simply
   * processes anaglyph combine pixel-for-pixel on the raw arrays.
   * @param lcol First screen colour.
   * @param rcol Second screen colour.
   * @param limg First image pixels as native pointer.
   * @param rimg Second image pixels as native pointer.
   * @param dest Store resulting values here, must be allocated already.
   */
  private static native void doCombine (int lcol, int rcol,
                                        long limg, long rimg, int[] dest);

  static
    {
      System.loadLibrary ("anacam");
    }

}
