/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package eu.domob.anacam;

import android.graphics.Bitmap;

/**
 * This represents a bitmap graphics just like android.graphics.Bitmap, but
 * optimised for performance and especially using as few copies as possible
 * for my purposes.  In addition, it allows to move the data to memory
 * allocated in native code, since this helps keep below the low
 * heap memory limit.  This native memory is then still usable to
 * combine anaglyphs in native code.
 */
public class MyBitmap
{

  /** Pixel data, always in RGB8888.  */
  private int[] data;

  /** Hold value for native pointer.  */
  private long nativeData;

  /** Width.  */
  public final int width;
  /** Height.  */
  public final int height;

  /**
   * Construct it from the given data.  No copy is made.
   * @param px Pixel data.
   * @param w Width.
   * @param h Height.
   */
  public MyBitmap (int[] px, int w, int h)
  {
    data = px;
    nativeData = 0;
    width = w;
    height = h;
  }

  /**
   * Construct it from an Android Bitmap.  This is used in order to
   * utilise the jpeg data we get from real pictures, not just the
   * previews.
   * @param bmp The Android Bitmap object.
   */
  public MyBitmap (Bitmap bmp)
  {
    width = bmp.getWidth ();
    height = bmp.getHeight ();
    nativeData = 0;

    data = new int[width * height];
    bmp.getPixels (data, 0, width, 0, 0, width, height);
  }

  /**
   * Get the pixel data.
   * @return Pixel data.
   */
  public int[] getData ()
  {
    assert (data != null);
    return data;
  }

  /**
   * Get native data pointer.
   * @return Pointer to native data.
   */
  public long getNativeData ()
  {
    return nativeData;
  }

  /**
   * Convert stored data to native memory.
   */
  public void toNative ()
  {
    /* Only do this if we have non-native data.  This prevents clearing the
       data out when this is called twice.  */
    if (data == null)
      return;

    if (nativeData != 0)
      freeNative (nativeData);

    nativeData = copyToNative (data);
    data = null;
  }

  /**
   * Convert the data to grayscale.  For this, the picture must not yet
   * be native!  We also need grayscale'd data in first-picture preview
   * where we want to display it directly.
   */
  public void toGrayscale ()
  {
    assert (data != null);
    doGrayscale (data);
  }

  /**
   * Free native pointer on finalization.
   */
  @Override
  protected void finalize ()
  {
    if (nativeData != 0)
      freeNative (nativeData);
  }

  /**
   * Native method to free a pointer.
   * @param ptr Pointer to free.
   */
  private static native void freeNative (long ptr);

  /**
   * Native method to allocate memory for the pixel data and store
   * it there.
   * @param data Pixels as Java array.
   * @return Pointer to allocated native memory.
   */
  private static native long copyToNative (int[] data);

  /**
   * Convert data in native code to grayscale (in place).
   * @param data Pixels as Java array.
   */
  private static native void doGrayscale (int[] data);

}
