/*
    AnaCam.  Take anaglyph pictures on Android.
    Copyright (C) 2012  Daniel Kraft <d@domob.eu>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

package eu.domob.anacam;

import android.content.Context;
import android.content.SharedPreferences;

import android.graphics.Bitmap;
import android.graphics.Canvas;
import android.graphics.Rect;

import android.hardware.Camera;

import android.os.Debug;

import android.preference.PreferenceManager;

import android.util.Log;

import android.view.SurfaceHolder;
import android.view.SurfaceView;

/**
 * SurfaceView used to do the camera preview including the anaglyph
 * modifications to it.
 */
public class Preview extends SurfaceView
  implements SurfaceHolder.Callback, Camera.PreviewCallback
{

  /** The camera handle.  */
  private Camera cam;

  /** Store SharedPreferences object.  */
  private SharedPreferences pref;

  /** The anaglyph combiner.  */
  private Anaglyph ana;
  /** The YUV decoder.  */
  private YuvDecoder yuvDecoder;

  /** Remember whether we have a first picture already.  */
  private boolean haveFirst;

  /** Width to resize preview images to.  */
  private int prevW;
  /** Height to resize preview images to.  */
  private int prevH;
  /** x position for drawing preview images.  */
  private int prevX;
  /** y position for drawing preview images.  */
  private int prevY;

  /** Store camera parameters for performance.  */
  private Camera.Parameters param;

  /**
   * Create it and register the callback to set up camera when we are ready.
   * @param c The context.
   */
  public Preview (Context c)
  {
    super (c);
    pref = PreferenceManager.getDefaultSharedPreferences (c);

    cam = null;
    param = null;

    ana = new Anaglyph (c);
    setFirstPicture (null);

    yuvDecoder = new YuvDecoder ();

    SurfaceHolder holder = getHolder ();
    holder.addCallback (this);
    holder.setType (SurfaceHolder.SURFACE_TYPE_NORMAL);
  }

  /**
   * Take a picture by the camera.
   * @param cb Callback to use when data is available.
   */
  public void takePicture (Camera.PictureCallback cb)
  {
    cam.takePicture (null, null, null, cb);
  }

  /**
   * Set first picture already taken.
   * @param f First picture to set.
   */
  public void setFirstPicture (Bitmap f)
  {
    if (f != null)
      {
        final Camera.Size size = param.getPreviewSize ();
        final Bitmap scaled = Bitmap.createScaledBitmap (f, size.width,
                                                         size.height, false);
        f.recycle ();

        final MyBitmap myScaled = new MyBitmap (scaled);
        if (wantGray ())
          myScaled.toGrayscale ();
        ana.setFirst (myScaled);
        scaled.recycle ();
      }
    else
      ana.setFirst (null);

    haveFirst = (f != null);
  }

  /**
   * Start preview.  This is used to re-start it after taking pictures.
   */
  public void start ()
  {
    if (AnaCam.PROFILING == AnaCam.PROFILE_SECOND)
      Debug.startMethodTracing ();
    cam.startPreview ();
  }

  /**
   * Query for the camera's jpeg quality.
   * @return Camera jpeg quality.
   */
  public int getJpegQuality ()
  {
    return param.getJpegQuality ();
  }

  /**
   * Free all unneeded buffers to make as much memory as possible available.
   */
  public void freeBuffers ()
  {
    ana.freeBuffers ();
    yuvDecoder.freeBuffers ();
  }

  /**
   * When the surface is created, set up the camera preview.
   * @param holder Surface holder.
   */
  public void surfaceCreated (SurfaceHolder holder)
  {
    assert (cam == null);
    cam = Camera.open ();
    cam.setPreviewCallback (this);

    param = cam.getParameters ();
    Log.i (AnaCam.TAG, String.format ("Preview is %dx%d.",
           param.getPreviewSize ().width, param.getPreviewSize ().height));
    Log.i (AnaCam.TAG,
           String.format ("Format is %d.", param.getPreviewFormat ()));
    Log.i (AnaCam.TAG,
           String.format ("Quality is %d.", param.getJpegQuality ()));

    if (AnaCam.PROFILING == AnaCam.PROFILE_FIRST)
      Debug.startMethodTracing ();
    cam.startPreview ();
  }

  /**
   * Clean up everything when the surface is destroyed.
   * @param holder Surface holder.
   */
  public void surfaceDestroyed (SurfaceHolder holder)
  {
    cam.setPreviewCallback (null);
    cam.stopPreview ();

    if (AnaCam.PROFILING == AnaCam.PROFILE_FIRST
        || AnaCam.PROFILING == AnaCam.PROFILE_SECOND)
      Debug.stopMethodTracing ();

    cam.release ();
    cam = null;
    param = null;
  }

  /**
   * Surface changed, remember new size and calculate position where we
   * want to draw the preview image.
   * @param holder Surface holder.
   * @param fmt Format.
   * @param w Width.
   * @param h Height.
   */
  public void surfaceChanged (SurfaceHolder holder, int fmt, int w, int h)
  {
    Log.i (AnaCam.TAG, String.format ("Display size is %dx%d.", w, h));

    final Camera.Size size = param.getPreviewSize ();
    final double fx = w / (float) size.width;
    final double fy = h / (float) size.height;
    final double f = Math.min (fx, fy);

    prevW = (int) (f * size.width);
    prevH = (int) (f * size.height);

    prevX = (w - prevW) / 2;
    prevY = (h - prevH) / 2;
  }

  /**
   * Handle camera preview frame.
   * @param data Data array.
   * @param c The camera that gave the data.
   */
  public void onPreviewFrame (byte[] data, Camera c)
  {
    assert (c == cam);
    final int w = param.getPreviewSize ().width;
    final int h = param.getPreviewSize ().height;

    final int fmt = param.getPreviewFormat ();
    yuvDecoder.setInput (data, w, h, fmt);
    MyBitmap bmp = yuvDecoder.decode (wantGray ());
    data = null;

    MyBitmap preview;
    if (haveFirst)
      {
        ana.setSecond (bmp);
        bmp = null;
        preview = ana.combine ();
        ana.setSecond (null);
      }
    else
      preview = bmp;

    Canvas canvas = null;
    try
      {
        canvas = getHolder ().lockCanvas ();
        final Rect dst = new Rect (prevX, prevY, prevX + prevW, prevY + prevH);

        /* Here we have to create an Android bitmap in order to do scaling
           when we draw it.  */
        final Bitmap previewBmp = Bitmap.createBitmap (preview.getData (), w, h,
                                                       Bitmap.Config.ARGB_8888);
        preview = null;
        canvas.drawBitmap (previewBmp, null, dst, null);
        previewBmp.recycle ();
      }
    finally
      {
        if (canvas != null)
          getHolder ().unlockCanvasAndPost (canvas);
      }
  }

  /**
   * Check the preferences to see whether we want grayscale preview.
   * @return True if yes.
   */
  private boolean wantGray ()
  {
    if (pref.getBoolean ("grayscale", false))
      return true;
    return pref.getBoolean ("grayPreview", false);
  }

}
